#define NAND_PAGE_SIZE		(is_cx?0x800:0x200)
#define MODEL_OFFSET		0x804
#define LOCAL_OFFSET		0x808
#define LOCAL_SIZE			8
#define ROM_OFFSET			0x82e
#define SDRAM_OFFSET		0x840
#define BOOTG_OFFSET		0x890

#define MANUF_PAGE_OFFSET	(is_cx?0x000:0x000)
#define BOOT2_PAGE_OFFSET	(is_cx?0x040:0x020)
#define BOOTD_PAGE_OFFSET	(is_cx?0x580:0xA80)
#define DIAGS_PAGE_OFFSET 	(is_cx?0x640:0xB00) // seems ok for TI-Nspire CM
#define DIAGD_PAGE_OFFSET	(is_cx?0x7A0:0xF80) // needs to be checked for TI-Nspire CM
#define DIAGE_PAGE_OFFSET	(is_cx?0x7C0:0x1000) // quick-dirty-hack for CM

#define MANUF_PAGES_NUM		(is_cx?0x040:0x020)
#define BOOT2_PAGES_NUM		(is_cx?0x540:0xA60)
#define BOOTD_PAGES_NUM		(is_cx?0x0C0:0x080)
#define DIAGS_PAGES_NUM		(is_cx?0x160:0x480)
#define DIAGD_PAGES_NUM		(is_cx?0x020:0x080) // filesystem seems to start lower on TI-Nspire CM
#define DIAGE_PAGES_NUM		(is_cx?0x040:0x000) // quick-dirty-hack for CM

#define MANUF_NAND_OFFSET	(MANUF_PAGE_OFFSET*NAND_PAGE_SIZE)
#define BOOT2_NAND_OFFSET	(BOOT2_PAGE_OFFSET*NAND_PAGE_SIZE)
#define BOOTD_NAND_OFFSET	(BOOTD_PAGE_OFFSET*NAND_PAGE_SIZE)
#define DIAGS_NAND_OFFSET	(DIAGS_PAGE_OFFSET*NAND_PAGE_SIZE)
#define DIAGD_NAND_OFFSET	(DIAGD_PAGE_OFFSET*NAND_PAGE_SIZE)
#define DIAGE_NAND_OFFSET	(DIAGE_PAGE_OFFSET*NAND_PAGE_SIZE)

#define MANUF_SIZE		(MANUF_PAGES_NUM * NAND_PAGE_SIZE)
#define BOOT2_SIZE		(BOOT2_PAGES_NUM * NAND_PAGE_SIZE)
#define BOOTD_SIZE		(BOOTD_PAGES_NUM * NAND_PAGE_SIZE)
#define DIAGS_SIZE		(DIAGS_PAGES_NUM * NAND_PAGE_SIZE)
#define DIAGD_SIZE		(DIAGD_PAGES_NUM * NAND_PAGE_SIZE)
#define DIAGE_SIZE		(DIAGE_PAGES_NUM * NAND_PAGE_SIZE)

#define MANUF_NAND_ENDOFFSET	(MANUF_NAND_OFFSET+MANUF_SIZE-1)
#define BOOT2_NAND_ENDOFFSET	(BOOT2_NAND_OFFSET+BOOT2_SIZE-1)
#define BOOTD_NAND_ENDOFFSET	(BOOTD_NAND_OFFSET+BOOTD_SIZE-1)
#define DIAGS_NAND_ENDOFFSET	(DIAGS_NAND_OFFSET+DIAGS_SIZE-1)
#define DIAGD_NAND_ENDOFFSET	(DIAGD_NAND_OFFSET+DIAGD_SIZE-1)
#define DIAGE_NAND_ENDOFFSET	(DIAGE_NAND_OFFSET+DIAGE_SIZE-1)

static const unsigned read_nand_addrs[]			= {0x10071F5C, 0x10071EC4, 0x10071658, 0X100715E8, 0x1006E0AC, 0x1006E03C}; // ns, cas, cx, cxcas, cm, cmcas
#define read_nand SYSCALL_CUSTOM(read_nand_addrs ,void,  void* dest, int size, int offset, int, int percent_max, void *progress_cb )

static const unsigned write_nand_addrs[]		= {0x10072298, 0x10072200, 0x10071994, 0x10071924, 0x1006E3E8, 0x1006E378}; // ns, cas, cx, cxcas, cm, cmcas
#define write_nand SYSCALL_CUSTOM(write_nand_addrs,int,  void *source,int size, unsigned int endoffset)

static const unsigned nand_erase_range_addrs[]	= {0x100724FC, 0x10072464, 0x10071BF8, 0x10071B88, 0x1006E64C, 0x1006E5DC}; // ns, cas, cx, cxcas, cm, cmcas
#define nand_erase_range SYSCALL_CUSTOM(nand_erase_range_addrs,int, int offset,int end)

#define NLANG 16
#define LANG_NONE	0
#define LANG_ZHCN	1
#define LANG_AR		2
#define LANG_FR		3
#define LANG_DA		4
#define LANG_DE		5
#define LANG_EN		6
#define LANG_ENGB	7
#define LANG_ES		8
#define LANG_FI		9
#define LANG_IT		10
#define LANG_NL		11
#define LANG_NLBE	12
#define LANG_NO		13
#define LANG_PT		14
#define LANG_SV		15
#define LANG_ZHTW	16

#define LOCAL_CHINESE	"zh_CN\x00\x00\x00"
#define LOCAL_INTERNA	"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"

int eraseFlash(int start, int end);
int writeFlash(char* data, int size, int start);
int getLocale(char* buf);