//--------- nsNandMgr.c ---------
// Requires Ndless 3.1
// Caution: This program has system routines capable of bricking the calculator, if used for other purposes.
// An inexpensive RS232-USB module is needed to recover from such an incident.

#include <os.h>
#include "console.h"
#include "screen.h"
#include "imgdump.h"
#include "nand.h"
#include "tools.h"

#define VERSION_SIZE 9
#define TYPE_SIZE 7
#define FLAGS_SIZE 8
#define ID_SIZE 3
#define SIZE_SIZE 4

extern unsigned char back[320*240*3];
unsigned char sscreen[320*240*2];
extern char locales[NLANG+1][LOCAL_SIZE];
extern char dlocales[NLANG+1][4];

void dispmodel(int id, int localid, int dmodel, int dlocal, int dtype)
{	switch(id)
	{	case 0x0C:
		case 0x0D:
		case 0x0E:
		case 0x0F:
		case 0x10:
		case 0x11:
		case 0x12:
			disp("Nspire",0,1);
			if(dmodel)
			{	switch(id)
				{	case 0x0F:
					case 0x10:
						disp(" CX",0,1);
						break;
					case 0x11:
					case 0x12:
						disp(" CM",0,1);
						break;
					case 0x0D:
						disp(" Lab",0,1);
						break;
				}
			}
			if(dlocal)
			{	if(localid>=0 && localid<=NLANG)
					disp(dlocales[localid],0,1);
			}
			if(dtype)
			{	switch(id)
				{	case 0x0C:
					case 0x0F:
					case 0x11:
						disp(" CAS",0,1);
						break;
				}
			}
			break;
	}
}

void dispmodelasic(int modeltype)
{	if(modeltype>=0 && modeltype<=3)
	{	int id=0;
		switch(modeltype)
		{	case 0:
				id=0x0E; // non-CAS
				break;
			case 1:
				id=0x0C; // CAS
				break;
			case 2:
				id=0x12; // CM non-CAS
				break;
			case 3:
				id=0x11; // CM CAS
				break;
		}
		dispmodel(id,0,1,0,1);
	}
}

void dispBack()
{	int ssize=320*240/2;
	if(is_cx) ssize=320*240*2;
	memcpy(SCREEN_BASE_ADDR,sscreen,ssize);
}

void derror(char* str)
{	dispBack();
	resetConsole();
	displn(str,1,1);
	disp("Press a key...",0,1);
	while(any_key_pressed());
	while(!any_key_pressed());
	dispBack();
}

int warning(char* str)
{	dispBack();
	resetConsole();
	displn(str,1,1);
	disp("Press y/n...",0,1);
	if(!waitConfirm())
	{	dispBack();
		return 0;
	}
	return 1;
}

void truncSpace(char* str, int size)
{	int i;
	for(i=0;i<size;i++)
	{	if(str[i]==' ')
			str[i]=0;
		if(str[i]==0)
			return;
	}
}

void getFilesDataInfo(char* data_buf, char* file, int size, char* data_type, char* data_version, int* data_size, int* compressed, int* fsize, int* id)
{	memset(data_buf,0,size);
	FILE* h = fopen(file,"rb");
	*fsize=0;
	if(h)
	{	*fsize=filelength(h);
		fread(data_buf,1,size,h);
		fclose(h);
	}
	memset(data_version,0,VERSION_SIZE);
	if(*data_buf) getRawField(data_buf,0x8020,VERSION_SIZE,data_version, size);
	truncSpace(data_version,VERSION_SIZE);
	memset(data_type,0,TYPE_SIZE);
	if(*data_buf) getRawField(data_buf,0x8040,TYPE_SIZE,data_type, size);
	truncSpace(data_type,TYPE_SIZE);
	char data_flags[FLAGS_SIZE];
	memset(data_flags,0,FLAGS_SIZE);
	if(*data_buf) getRawField(data_buf,0x8080,FLAGS_SIZE,data_flags, size);
	*compressed = data_flags[FLAGS_SIZE-1];
	*data_size=0;
	if(*data_buf) *data_size=getFieldSize(data_buf,0x8070, size);
	char data_id[ID_SIZE];
	memset(data_id,0,ID_SIZE);
	if(*data_buf) getRawField(data_buf,0x8010,ID_SIZE,data_id, size);
	*id=chex2int(data_id[1])*16+chex2int(data_id[2]);
}


void getFlashDataInfo(char* data_buf, int nand_offset, int size, char* data_type, char* data_version, int* data_size, int* compressed, int* id)
{	memset(data_buf,0,size);
	read_nand(data_buf,size,nand_offset,0,0,NULL);
	memset(data_version,0,VERSION_SIZE);
	if(*data_buf) getRawField(data_buf,0x8020,VERSION_SIZE,data_version, size);
	truncSpace(data_version,VERSION_SIZE);
	memset(data_type,0,TYPE_SIZE);
	if(*data_buf) getRawField(data_buf,0x8040,TYPE_SIZE,data_type, size);
	truncSpace(data_type,TYPE_SIZE);
	char data_flags[FLAGS_SIZE];
	memset(data_flags,0,FLAGS_SIZE);
	if(*data_buf) getRawField(data_buf,0x8080,FLAGS_SIZE,data_flags, size);
	*compressed = data_flags[FLAGS_SIZE-1];
	*data_size=0;
	if(*data_buf) *data_size=getFieldSize(data_buf,0x8070, size);
	char data_id[ID_SIZE];
	memset(data_id,0,ID_SIZE);
	if(*data_buf) getRawField(data_buf,0x8010,ID_SIZE,data_id, size);
	*id=chex2int(data_id[1])*16+chex2int(data_id[2]);
}

void dispManufInfo(int id, int localid, int ram, int rom)
{	int ccol = getCol();
	switch(id)
	{	case 0x0C:
		case 0x0D:
		case 0x0E:
		case 0x0F:
		case 0x10:
		case 0x11:
		case 0x12:
			dispmodel(id,localid,1,1,1);
			displn("",0,1);
			setCol(ccol);
			dispi(ram,0,1);
			disp("/",0,1);
			dispi(rom,0,1);
			disp("MB RAM/ROM",0,1);
		default:
			displn("",0,1);
	}
}

void dispDataInfo(char* data_type, char* data_version, int data_compressed)
{	int ccol = getCol();
	disp(data_type,0,1);
	setCol(ccol+TYPE_SIZE-1);
	disp(data_version,0,1);
	if(data_compressed)	disp("c",0,1);
}

int isCas(int modeltype)
{	return modeltype>=4 || modeltype==1 || modeltype==3;
}

int isNCas(int modeltype)
{	return modeltype>=4 || modeltype==0 || modeltype==2;
}

int isValidManuf(int modelid, int newid, int modeltype)
{	return	((modelid==0x0E || modelid==0x0C) && newid==0x0C && isCas(modeltype))		// Nspire -> Nspire CAS
			|| ((modelid==0x0C || modelid==0x0E) && newid==0x0E && isNCas(modeltype))	// Nspire CAS -> Nspire
			|| ((modelid==0x10 || modelid==0x0F) && newid==0x0F && isCas(modeltype))	// Nspire CX -> Nspire CX CAS
			|| ((modelid==0x0F || modelid==0x10) && newid==0x10 && isNCas(modeltype))	// Nspire CX CAS -> Nspire CX
			|| ((modelid==0x12 || modelid==0x11) && newid==0x11 && isCas(modeltype))	// Nspire CM -> Nspire CM CAS
			|| ((modelid==0x11 || modelid==0x12) && newid==0x12 && isNCas(modeltype))	// Nspire CM CAS -> Nspire CM
			|| (modelid==0x0D && newid==0x0D)
			;
}

int isValidManufData(int modelid, int id)
{	return	((modelid==0x0E || modelid==0x0C || modelid==0x0D) && (id==0x0E || id==0x0C || id==0x0D))
			|| ((modelid==0x0F || modelid==0x10) && (id==0x0F || id==0x10))
			|| ((modelid==0x11 || modelid==0x12) && (id==0x11 || id==0x12));
}

int isValidFlashBoot2(int modelid, int id, char* type)
{	if(strcmp(type,"BOOT2"))
		return 0;
	return	((modelid==0x0E || modelid==0x0C) && (id==0x0E || id==0x0C))
			|| ((modelid==0x0F || modelid==0x10) && (id==0x0F || id==0x10))
			|| ((modelid==0x11 || modelid==0x12) && (id==0x11 || id==0x12))
			|| (modelid==0x0D && id==0x0D);
}

int isValidFlashDiags(int modelid, int id,char* type)
{	return isValidFlashBoot2(modelid,id,"BOOT2");
}

int isValidRun(int modelid, int id, char* type)
{	if(!strcmp(type,"BOOT2"))
		return isValidFlashBoot2(modelid,id,type);
	return	((modelid==0x0E || modelid==0x0C || modelid==0x0D) && (id==0x0E || id==0x0C || id==0x0D))
			|| ((modelid==0x0F || modelid==0x10 || modelid==0x11 || modelid==0x12) && (id==0x0F || id==0x10 || id==0x11 || id==0x12));
}

int main(void) {

	char* manufflashdata=malloc(MANUF_SIZE);
	char* manuffilesdata=malloc(MANUF_SIZE);
	int manuffilessize;
	
	char* boot2flashdata=malloc(BOOT2_SIZE);
	char boot2flashversion[VERSION_SIZE];
	char boot2flashtype[TYPE_SIZE];

	char* boot2filesdata=malloc(BOOT2_SIZE);
	char boot2filesversion[VERSION_SIZE];
	char boot2filestype[TYPE_SIZE];
	int boot2filessize;

	char* diagsflashdata=malloc(DIAGS_SIZE+DIAGD_SIZE+DIAGE_SIZE);
	char diagsflashversion[VERSION_SIZE];
	char diagsflashtype[TYPE_SIZE];

	char* diagsfilesdata=malloc(DIAGS_SIZE+DIAGD_SIZE+DIAGE_SIZE);
	char diagsfilesversion[VERSION_SIZE];
	char diagsfilestype[TYPE_SIZE];
	int diagsfilessize;

	char downgradeversion[VERSION_SIZE];

	char* bootddata=malloc(BOOTD_SIZE);					

	initScr();
	dispBufImgRGB(sscreen,0,0,back,320,240);
	dispBack();
	int menu=0;
	int option=0;
	char* fmode;
	char* file;
	int error=0;
	char* flashbuf;
	char* filesbuf;
	int size, offset, endoffset, filessize;
	int result;
	int i;
	int bootorder;
	int boot2flashcompressed,diagsflashcompressed,boot2filescompressed,diagsfilescompressed,boot2flashdsize,diagsflashdsize,boot2filesdsize,diagsfilesdsize;
	int boot2flashid;
	int diagsflashid;
	int boot2filesid;
	int diagsfilesid;
	int manufflashid;
	int manufflashloca;
	int manufflashram;
	int manufflashrom;
	int manuffilesid;
	int manuffilesloca;
	int manuffilesram;
	int manuffilesrom;
	unsigned int modeltype = *(volatile unsigned int*) 0x900A002C;
	modeltype = (modeltype>>(58-32))&(0b11111);
	char* flashtype;
	int flashid;
	int update=1;
	
	while(menu>=0)
	{	if(update)
		{	read_nand(manufflashdata,MANUF_SIZE,MANUF_NAND_OFFSET,0,0,NULL);
			getFlashDataInfo(boot2flashdata,BOOT2_NAND_OFFSET,BOOT2_SIZE,boot2flashtype,boot2flashversion,&boot2flashdsize,&boot2flashcompressed,&boot2flashid);
			getFlashDataInfo(diagsflashdata,DIAGS_NAND_OFFSET,DIAGS_SIZE+DIAGD_SIZE+DIAGE_SIZE,diagsflashtype,diagsflashversion,&diagsflashdsize,&diagsflashcompressed,&diagsflashid);
			getFilesDataInfo(boot2filesdata,"/documents/ndless/boot2.img.tns",BOOT2_SIZE,boot2filestype,boot2filesversion,&boot2filesdsize,&boot2filescompressed,&boot2filessize,&boot2filesid);
			getFilesDataInfo(diagsfilesdata,"/documents/ndless/diags.img.tns",DIAGS_SIZE+DIAGD_SIZE+DIAGE_SIZE,diagsfilestype,diagsfilesversion,&diagsfilesdsize,&diagsfilescompressed,&diagsfilessize,&diagsfilesid);
			read_nand(bootddata,BOOTD_SIZE,BOOTD_NAND_OFFSET,0,0,NULL);
			memset(manuffilesdata,0,MANUF_SIZE);
			FILE* h = fopen("/documents/ndless/manuf.img.tns","rb");
			manuffilessize=0;
			if(h)
			{	manuffilessize=filelength(h);
				fread(manuffilesdata,1,MANUF_SIZE,h);
				fclose(h);
			}
			bootorder=1;
			*downgradeversion=0;
			for(i=0;i<BOOTD_PAGES_NUM;i++)
			{	char* tmp=bootddata+i*NAND_PAGE_SIZE;
				char tmp2[5];
				memcpy(tmp2,tmp,4);
				tmp2[4]=0;
				if(!strcmp(tmp2,"\xAA\xC6\x8C\x92"))
				{	sprintf(downgradeversion,"%i.%i.%i",tmp[7],tmp[6],tmp[5]*256+tmp[4]);
					bootorder=tmp[0x10]|tmp[0x11]|tmp[0x12]|tmp[0x13];
				}
			}		
			manufflashid = manufflashdata[MODEL_OFFSET];
			manufflashloca = getLocale(manufflashdata+LOCAL_OFFSET);
			if(manufflashdata[SDRAM_OFFSET]==0x12)	manufflashram=64;
			else									manufflashram=32;
			if(manufflashdata[ROM_OFFSET]==0x32)	manufflashrom=128;
			else									manufflashrom=32;
			manuffilesid = manuffilesdata[MODEL_OFFSET];
			manuffilesloca = getLocale(manuffilesdata+LOCAL_OFFSET);
			if(manuffilesdata[SDRAM_OFFSET]==0x12)	manuffilesram=64;
			else									manuffilesram=32;
			if(manuffilesdata[ROM_OFFSET]==0x32)	manuffilesrom=128;
			else									manuffilesrom=32;
			update=0;
		}
		while(!option && menu>=0)
		{	dispBack();
			resetConsole();
			dispmodelasic(modeltype);
			setCol(16);
							displn("Flash  |FileSystem",0,1);
			displn("-----------------------+----------------",0,1);
			disp("Manuf: ",0,1);
			int ccol = getCol();
			int cline = getLine();
			dispManufInfo(manufflashid,getLocale(manufflashdata+LOCAL_OFFSET),manufflashram,manufflashrom);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			setLine(cline);
			displn("|",0,1);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			disp("|",0,1);
			setLine(cline);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE+1);
			dispManufInfo(manuffilesid,getLocale(manuffilesdata+LOCAL_OFFSET),manuffilesram,manuffilesrom);
			setLine(cline+1);
			displn("",0,1);
			displn("-----------------------+----------------",0,1);
			disp("Boot2: ",0,1);
			dispDataInfo(boot2flashtype,boot2flashversion,boot2flashcompressed);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			disp("|",0,1);
			dispDataInfo(boot2filestype,boot2filesversion,boot2filescompressed);
			displn("",0,1);
			disp("       ",0,1);
			dispmodel(boot2flashid,0,1,0,0);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			disp("|",0,1);
			dispmodel(boot2filesid,0,1,0,0);
			displn("",0,1);
			displn("-----------------------+---------------",0,1);
			disp("Diags: ",0,1);
			dispDataInfo(diagsflashtype,diagsflashversion,diagsflashcompressed);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			disp("|",0,1);
			dispDataInfo(diagsfilestype,diagsfilesversion,diagsfilescompressed);
			displn("",0,1);
			disp("       ",0,1);
			dispmodel(diagsflashid,0,1,0,0);
			setCol(ccol+TYPE_SIZE+VERSION_SIZE);
			disp("|",0,1);
			dispmodel(diagsfilesid,0,1,0,0);
			displn("",0,1);
			displn("-----------------------+----------------",0,1);
			disp("BootData: ",0,1);
			disp("minimal OS = ",0,1);
			displn(downgradeversion,0,1);
			disp("          boot order = ",0,1);
			if(bootorder)	disp("Diags Boot2",0,1);
			else			disp("Boot2 Diags",0,1);
			displn("",0,1);
			if(menu==0)
			{	displn("1) Manuf code  4) Manuf Data",0,1);
				displn("2) Boot2 code  5) Boot2 Data",0,1);
				displn("3) Diags code  6) Diags Data",0,1);
			}
			else if(menu==1)
			{		displn(" 1) Nspire  7) CM CAS  D) ar  J) fr",0,1);
					displn(" 2) CAS     8) Intern  E) da  K) it",0,1);
					displn(" 3) Lab     9) -C      F) de  L) nl",0,1);
					displn(" 4) CX      A) en_GB   G) en  M) no",0,1);
					displn(" 5) CX CAS  B) nl_BE   H) es  N) pt",0,1);
					displn(" 6) CM      C) zh_TW   I) fi  O) sv",0,1);
				}
			else if(menu==2)
			{	displn(" 1) Dump  Boot2:   Flash  -> Filesystem",0,1);
				displn(" 2) Flash Boot2:   Flash <-  Filesystem",0,1);
				displn(" 3) Run   Boot2 in Filesystem",0,1);
			}
			else if(menu==3)
			{	displn(" 1) Dump  Diags:   Flash  -> Filesystem",0,1);
				displn(" 2) Flash Diags:   Flash <-  Filesystem",0,1);
				displn(" 3) Run   Diags in Filesystem",0,1);
			}
			else if(menu==4)
			{	displn(" 1) Dump  ManufD:   Flash  -> Filesystem",0,1);
				displn(" 2) Flash ManufD:   Flash <-  Filesystem",0,1);
			}
			else if(menu==5)
			{	displn(" 4) Erase BootD",0,1);
				displn(" 5) Remove downgrade protection",0,1);
				displn(" 6) Switch boot order",0,1);
			}
			else if(menu==6)
			{	displn(" 4) Erase Diags Data",0,1);
			}
			if(menu>0)
				disp(" 0) Back",0,1);
			else
				disp("0) Exit",0,1);
			while(!any_key_pressed());
			if(isKeyPressed(KEY_NSPIRE_ESC) || isKeyPressed(KEY_NSPIRE_0))
			{	if(menu>0) menu=0;
				else	   menu=-1;
			}
			else if(menu==0)
			{	if(isKeyPressed(KEY_NSPIRE_1)) menu=1;
				if(isKeyPressed(KEY_NSPIRE_2)) menu=2;
				if(isKeyPressed(KEY_NSPIRE_3)) menu=3;
				if(isKeyPressed(KEY_NSPIRE_4)) menu=4;
				if(isKeyPressed(KEY_NSPIRE_5)) menu=5;
				if(isKeyPressed(KEY_NSPIRE_6)) menu=6;
			}
			else if(menu>=0)
			{	if(isKeyPressed(KEY_NSPIRE_1)) option=1;
				if(isKeyPressed(KEY_NSPIRE_2)) option=2;
				if(isKeyPressed(KEY_NSPIRE_3)) option=3;
				if(isKeyPressed(KEY_NSPIRE_4)) option=4;
				if(isKeyPressed(KEY_NSPIRE_5)) option=5;
				if(isKeyPressed(KEY_NSPIRE_6)) option=6;
				if(isKeyPressed(KEY_NSPIRE_7)) option=7;
				if(isKeyPressed(KEY_NSPIRE_8)) option=8;
				if(isKeyPressed(KEY_NSPIRE_9)) option=9;
				if(isKeyPressed(KEY_NSPIRE_A)) option=10;
				if(isKeyPressed(KEY_NSPIRE_B)) option=11;
				if(isKeyPressed(KEY_NSPIRE_C)) option=12;
				if(isKeyPressed(KEY_NSPIRE_D)) option=13;
				if(isKeyPressed(KEY_NSPIRE_E)) option=14;
				if(isKeyPressed(KEY_NSPIRE_F)) option=15;
				if(isKeyPressed(KEY_NSPIRE_G)) option=16;
				if(isKeyPressed(KEY_NSPIRE_H)) option=17;
				if(isKeyPressed(KEY_NSPIRE_I)) option=18;
				if(isKeyPressed(KEY_NSPIRE_J)) option=19;
				if(isKeyPressed(KEY_NSPIRE_K)) option=20;
				if(isKeyPressed(KEY_NSPIRE_L)) option=21;
				if(isKeyPressed(KEY_NSPIRE_M)) option=22;
				if(isKeyPressed(KEY_NSPIRE_N)) option=23;
				if(isKeyPressed(KEY_NSPIRE_O)) option=24;
			}
			while(any_key_pressed());
		}

		if(menu>=2 && menu<=4)
		{	if(menu==2) file="/documents/ndless/boot2.img.tns";
			else if(menu==3) file="/documents/ndless/diags.img.tns";
			else if(menu==4) file="/documents/ndless/manuf.img.tns";
		}
				
		// MANUF/BOOT2/DIAGS/BOOTD/DIAGD
		if(	(menu==1 && option>=1) || (menu==2 && (option>=1 && option<=3)) || (menu==3 && (option>=1 && option<=3))
			|| (menu==4 && (option>=1 && option<=2)) || (menu==6 && option==4) || (menu==5 && (option>=4 && option<=6)))
		{	update=1;
			int extsize=0;
			int ext2size=0;
			int flashdsize=0;
			int filesdsize=0;
			FILE* hfile;
			if(menu==2)			{ flashbuf=boot2flashdata;	size=BOOT2_SIZE; offset=BOOT2_NAND_OFFSET; endoffset=BOOT2_NAND_ENDOFFSET; filesbuf=boot2filesdata; filessize=boot2filessize;	flashid=boot2filesid;		flashtype=boot2filestype; flashdsize=boot2flashdsize; filesdsize=boot2filesdsize;}
			else if(menu==3)	{ flashbuf=diagsflashdata;	size=DIAGS_SIZE; offset=DIAGS_NAND_OFFSET; endoffset=DIAGS_NAND_ENDOFFSET; filesbuf=diagsfilesdata; filessize=diagsfilessize;	flashid=diagsfilesid;		flashtype=diagsfilestype; flashdsize=diagsflashdsize; filesdsize=diagsfilesdsize; extsize=DIAGD_SIZE; ext2size=DIAGE_SIZE; if(manufflashid==0x11 || manufflashid==0x12) ext2size=0;}
			else if(menu==5)	{ 							size=BOOTD_SIZE; offset=BOOTD_NAND_OFFSET; endoffset=BOOTD_NAND_ENDOFFSET; filesbuf=bootddata;		filessize=BOOTD_SIZE;		flashid=manufflashid;		flashtype="BOOT2";		  }
			else if(menu==1)	{ flashbuf = manufflashdata;size=MANUF_SIZE; offset=MANUF_NAND_OFFSET; endoffset=MANUF_NAND_ENDOFFSET; 							filessize=MANUF_SIZE;	flashid=manufflashid;		flashtype="BOOT2";			}
			else if(menu==4)	{ flashbuf = manufflashdata;size=MANUF_SIZE; offset=MANUF_NAND_OFFSET; endoffset=MANUF_NAND_ENDOFFSET; filesbuf=manuffilesdata; filessize=manuffilessize;	flashid=manuffilesid;		flashtype="DIAGS";			}
			else if(menu==6)	{ 							size=DIAGD_SIZE; offset=DIAGD_NAND_OFFSET; endoffset=DIAGD_NAND_ENDOFFSET;							filessize=DIAGD_SIZE;		flashid=manufflashid;		flashtype="BOOT2";		  if(manufflashid==0x10 || manufflashid==0x0F) {size+=DIAGE_SIZE; endoffset+=DIAGE_SIZE;}}

			if(menu==1) // MANUF
			{	int code = -1;
				switch(option)
				{	case 1:
						code=0x0E;
						break;
					case 2:
						code=0x0C;
						break;
					case 3:
						code=0x0D;
						break;
					case 4:
						code=0x10;
						break;
					case 5:
						code=0x0F;
						break;
					case 6:
						code=0x12;
						break;
					case 7:
						code=0x11;
						break;
					case 8:
						manufflashloca=LANG_NONE;
						break;
					case 9:
						manufflashloca=LANG_ZHCN;
						break;
					case 10:
						manufflashloca=LANG_ENGB;
						break;
					case 11:
						manufflashloca=LANG_NLBE;
						break;
					case 12:
						manufflashloca=LANG_ZHTW;
						break;
					case 13:
						manufflashloca=LANG_AR;
						break;
					case 14:
						manufflashloca=LANG_DA;
						break;
					case 15:
						manufflashloca=LANG_DE;
						break;
					case 16:
						manufflashloca=LANG_EN;
						break;
					case 17:
						manufflashloca=LANG_ES;
						break;
					case 18:
						manufflashloca=LANG_FI;
						break;
					case 19:
						manufflashloca=LANG_FR;
						break;
					case 20:
						manufflashloca=LANG_IT;
						break;
					case 21:
						manufflashloca=LANG_NL;
						break;
					case 22:
						manufflashloca=LANG_NO;
						break;
					case 23:
						manufflashloca=LANG_PT;
						break;
					case 24:
						manufflashloca=LANG_SV;
						break;
					default:
						break;
				}
				if(option>=1 && option<=7)
				{	filesbuf = flashbuf;
					filesbuf[MODEL_OFFSET]=code;
					option=2;
					flashid=code;
				}
				if(option>=8 && option<=24)
				{	memcpy(flashbuf+LOCAL_OFFSET,locales[manufflashloca],LOCAL_SIZE);
					filesbuf = flashbuf;
					option=2;
					flashid=manufflashid;
				}
			}
			else if(menu==4 && (option==2 || option==4)) // FLASH/ERASE MANUF
			{	memcpy(filesbuf,flashbuf,BOOTG_OFFSET);
			}
			else if(menu==5 && (option==5 || option==6))		// BOOTD: remove downgrade protection
			{	int found = 0;
				for(i=0;i<BOOTD_PAGES_NUM;i++)
				{	char* tmp=bootddata+i*NAND_PAGE_SIZE;
					char tmp2[5];
					memcpy(tmp2,tmp,4);
					tmp2[4]=0;
					if(!strcmp(tmp2,"\xAA\xC6\x8C\x92"))
					{	found = 1;
						if(option==5)	tmp[7]=tmp[6]=tmp[5]=tmp[4]=0;
						else
						{	tmp[0x13]=tmp[0x12]=tmp[0x11]=0;
							if(bootorder)	tmp[0x10]=0;
							else			tmp[0x10]=1;
						}
					}
				}
				if(!found)		derror("Boot Data invalid: reset and try again.");
				else			option=2;
			}
			if((menu==2 || menu==3 || menu==4) && (option==2 || option==3))
				if(!*filesbuf)
				{	error=1;
					char* tmp="I/O error on file \0                              ";
					strcat(tmp,file);
					derror( tmp );
					error=1;
				}
			if(option==3 && !error)					// run BOOT2/DIAGS
			{	if(!isValidRun(manufflashid,flashid,flashtype))
				{	derror("Image rejected as determined unsafe if run on this model" );
					error=1;
				}
				if(!error)
				{	FILE* tfile;
					int s;
					char* rawfilesdata = 0;
					s = decompressFiles(filesbuf,size,&rawfilesdata);
					char* safe = 0;
					while((safe<0x11800000+s && safe>0x11800000-s) || !safe)
					{	safe = malloc(s);
					}
					memcpy(safe,rawfilesdata,s);
					memcpy((void*)0x11800000,safe,s);
					unsigned intmask = TCT_Local_Control_Interrupts(0);
					asm(
						".arm \n"
						"ldr pc, =0x11800000 @ jump to boot2\n"
					);
					// should never reach this
				}
			}
			else if(option==1)				// DUMP BOOT2/DIAGS
			{	hfile = fopen(file,"w+b");
				if(!hfile)
				{	error=1;
					derror("Error opening file" );
				}
				else
				{	if(flashdsize>size)
					{	size+=extsize;
						endoffset+=extsize;
						if(flashdsize>size)
						{	size+=ext2size;
							endoffset+=ext2size;
							if(flashdsize>size)
							{	if(!warning("WARNING: Dumped image seems corrupt. Do you want to continue?"))
									error=1;
							}
						}
					}
					if(size!=fwrite(flashbuf,1,size,hfile))
					{	error=1;
						derror("Error writing file" );
					}
					fclose(hfile);
				}
			}
			else if((option==2 || option==4) && !error)	// FLASH/ERASE BOOT2/DIAGS/BOOTD/DIAGD/MANUFD
			{	if(menu==1)
					if(!isValidManuf(manufflashid,flashid,modeltype))
						error=1;
				if(menu==4)
					if(!isValidManufData(manufflashid,flashid))
						error=1;
				if(menu==2)
					if(!isValidFlashBoot2(manufflashid,flashid,flashtype))
						error=1;
				if(menu==3)
					if(!isValidFlashDiags(manufflashid,flashid,flashtype))
						error=1;
				if(error)
					derror("Image rejected as determined unsafe if flashed on this model");
				if(!error)
					if(filesdsize>size)
					{	size+=extsize;
						endoffset+=extsize;
						if(filesdsize>size)
						{	size+=ext2size;
							endoffset+=ext2size;
							if(filesdsize>size)
							{	derror("Flashed image exceeds area and has been rejected");
								error=1;
							}
							else if(filessize>size)
							{	if(!warning("WARNING: Flashed image exceeds area size and will be truncated to valid code. Are you sure you want to proceed?"))
									error=1;
							}
						}
					}
				if(!error)
				{	error = eraseFlash(offset,endoffset);
					if(error)
						derror("Error erasing NAND" );
					else if(option==2)			// FLASH BOOT2/DIAGS
					{	error = writeFlash(filesbuf,size,offset);
						if(error)
							derror("Error writing to NAND" );
					}
				}
			}
		}
		option=0;
		error=0;
	}
	free(boot2flashdata);
	free(diagsflashdata);
	free(boot2filesdata);
	free(diagsfilesdata);
        return 0;                         
}
